(* $Id: godi_tty.ml 403 2004-05-02 13:26:31Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Term

type key =
  [ `Return
  | `BS
  | `ESC
  | `PageUp
  | `PageDown
  | `Beginning
  | `End
  | `Char of char
  | `None
  ]

type termcap =
    { init : string;
        (* INitialization string *)
      home : string;
        (* Set cursor to upper left corner *)
      clear : string;
        (* Clear screen, cursor at top left corner *)
      clear_eol : string;
        (* Clear until end of line *)
      to_bol : string;
        (* Go to beginning of current line *)
      to_bol_next : string;
        (* Go to beginnning of next line *)
      lines : int;
      columns : int;
      keys : (string * key) list;
        (* Only non-`Char keys *)
    };;


let try_keys l =
  let r =
  List.fold_left
    (fun acc (tcap, tinfo, key) ->
       try
	 let r = (convert(getstr tcap tinfo), key) :: acc in
	 (* prerr_endline ("Found key " ^ tinfo); *)
	 r
       with
	   Not_found ->
	     acc)
    []
    l
  in
  (* read_line(); *)
  r
;;


let find_key term seq =
  List.fold_left
    (fun acc (def_seq, key) ->
       if acc = `Not_found then
	 if (String.length seq < String.length def_seq) &&
	    (String.sub def_seq 0 (String.length seq) = seq) then
	   `Prefix
	 else
	   if seq = def_seq then
	     `Key key
	   else
	     acc
       else
	 acc
    )
    `Not_found
    term.keys
;;


let dumb_term =
  { init = "";
    home = "\027[H";
    clear = "\027[H\027[2J";
    clear_eol = "\027[K";
    to_bol = "\013";
    to_bol_next = "\013\010";
    lines = 24;
    columns = 80;
    keys = [ "\013", `Return;
	     "\010", `Return;
	     "\008", `BS;
	     "\127", `BS;
	   ];
  };;


let this_term =
  lazy (
    setup();
    try
      { init = 
	  (try convert(getstr "ks" "smkx") with Not_found -> "");
	home = convert(getstr "ho" "home");
	clear = convert(getstr "cl" "clear");
	clear_eol = convert(getstr "ce" "el");
	to_bol = convert(getstr "cr" "cr");
	to_bol_next = convert(getstr "cr" "cr") ^ convert(getstr "do" "cud1");
	lines = lines();
	columns = columns();
	keys = try_keys [ "kb", "kbs", `BS;
			  "kD", "kdch1", `BS;
			  "", "kent", `Return;
			  "kN", "knp", `PageDown;
			  "kP", "kpp", `PageUp;
			  "kd", "kcud1", `PageDown;
			  "ku", "kcuu1", `PageUp;
			  "kF", "kind", `PageDown;
			  "kR", "kri", `PageUp;
			  "%5", "knxt", `PageDown;
			  "%8", "kprv", `PageUp;
			  "kh", "khome", `Beginning;
			  "kH", "kll", `End;
			  "@7", "kend", `End;
			] @
	       [ "\013", `Return;
		 "\010", `Return; ];
      }
    with
	Not_found -> dumb_term
  )
;;



let get_terminal_capabilities() = Lazy.force this_term;;



let rec restart ?(onsignal = fun () -> ()) f arg =
  try
    f arg
  with
      Unix.Unix_error(Unix.EINTR,_,_) ->
	onsignal();
	restart f arg
;;


let last_sigcont_event = ref (-1.0) ;;
(* The Sys.time when the last SIGCONT signal was caught *)

let _ = Godi_exec.register_signal_handler
	  Sys.sigcont
	  (fun _ -> last_sigcont_event := Sys.time()) in
();;


let forced_refresh_event() = !last_sigcont_event;;


let poll_keyboard_tty ?(wait=false) () =
  let term = Lazy.force this_term in
  let tc = restart Unix.tcgetattr Unix.stdin in
  restart (Unix.tcsetattr Unix.stdin Unix.TCSANOW)
    { tc with
	Unix.c_ixon = false;
	Unix.c_icanon = false;
	Unix.c_echo = false;
	Unix.c_vmin = 1;
	Unix.c_vtime = 0;
    };
  try
    let onsignal = fun () -> raise Exit in
    (* When a signal happens: Force that the key `None is reported.
	     * This makes life easier when the user presses CTRL-Z followed
     * by fg and the screen must be repainted.
     *)
    let do_read =
      wait || (
	let rd, _, _ =
	  restart ~onsignal (Unix.select [Unix.stdin] [] []) 0.0 in
	rd <> []
      ) in
    if do_read then (
      let key = ref `None in
      let s = String.create 1 in
      let n = restart ~onsignal (Unix.read Unix.stdin s 0) 1 in
      assert(n = 1);    (* nobody must close stdin! *)
      (* Special handling of ESC: *)
      if s.[0] = '\027' then (
	let rd, _, _ =
	  restart ~onsignal (Unix.select [Unix.stdin] [] []) 0.1 in
	if rd <> [] then (
	  (* try to interpret an ESC sequence *)
	  let seq = ref (String.copy s) in
	  let state = ref `Prefix in
	  while !state = `Prefix do
	    ignore(restart ~onsignal (Unix.read Unix.stdin s 0) 1);
	    seq := !seq ^ s;
	    state := find_key term !seq
	  done;
	  match !state with
	      `Key k -> key := k
	    | `Not_found -> key := `None
	    | _ -> ()
	)
	else
	  key := `ESC;
      )
      else (
	(* look the key up: *)
	try
	  key := List.assoc s term.keys
	with
	    Not_found ->
	      key := `Char s.[0];
      );
      restart (Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH) tc;
      !key
    ) else (
      restart (Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH) tc;
      `None
    )
  with
    | Exit ->
	restart(Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH) tc;
	`None
    | error ->
	restart(Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH) tc;
	raise error
;;


let script = ref [] ;;

let poll_keyboard ?wait () =
  match !script with
      key :: script' ->
	script := script';
	key
    | [] ->
	poll_keyboard_tty ?wait ()
;;


let load_script filename =
  let f = open_in filename in
  let s = ref [] in
  try
    while true do
      let line = input_line f in
      match line with
	  "RETURN"    -> s := `Return :: !s
	| "BS"        -> s := `BS :: !s
	| "ESC"       -> s := `ESC :: !s
	| "PGUP"      -> s := `PageUp :: !s
	| "PGDOWN"    -> s := `PageDown :: !s
	| "BEGINNING" -> s := `Beginning :: !s
	| "END"       -> s := `End :: !s
	| _           -> for k = 0 to String.length line - 1 do
	                   s := (`Char line.[k]) :: !s
			 done
    done;
    assert false
  with
      End_of_file ->
	script := List.rev !s;
	close_in f
    | error ->
	close_in f;
	raise error
;;



